package org.leedberg.cotiro;

import java.io.IOException;

import org.apache.xmlbeans.XmlException;
import org.leedberg.cotiro.CotiroDocument.Cotiro;

/*
Cotiro 0.6.5
College Time and Room Scheduler Using Constraint Satisfaction
Copyright (C) 2003 - 2007, Gregory G. Leedberg

This program is free software; you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation; either version 2 of the License, or
(at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program; if not, write to the Free Software
Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
*/
public class CotiroConsole {

	/**
	 * @param args
	 */
	public static void main(String[] args) {
		boolean doConstraints = true;
		boolean doConsistency = true;
		boolean doHeuristics = true;
		boolean doSoft = true;
		boolean showStates = false;
		boolean consoleOutput = true;
		boolean overwrite = false;
		String outputName = null;
		String inputName = null;
		Cotiro cotiroInput = null;

		CotiroLogic.println(CotiroLogic.version);
		CotiroLogic.println("");

		if (args.length == 0) {
			CotiroLogic.println("Arguments:");
			CotiroLogic.println("--noconstraints\t\tDon't use constraints");
			CotiroLogic.println("--noconsistency\t\tDon't force arc consistency");
			CotiroLogic.println("--noheuristics\t\tDon't use heuristics for variable and value ordering");
			CotiroLogic.println("--nosoft\t\tTreat all constraints as hard");
			CotiroLogic.println("--showstates\t\tShow all visited states during the backtracking search");
			CotiroLogic.println("--noconsoleoutput\tDon't print the final schedule to the screen");
			CotiroLogic.println("--overwite\t\tIf output files already exist, overwrite them");
			CotiroLogic.println("--output [prefix]\tPrefix for for output file names.  Files will be named\n\t\t\tschedules\\[prefix].xml, .html, and .txt.");
			CotiroLogic.println("--input [filename]\tName of the input file, assumed to be in the\n\t\t\tdata\\ directory");

			return;
		}
		// Take an argument specifying the output file
		// From that, produce a .xml, .txt, and .html file
		// If any of these files already exist, exit unless --overwrite was used
		// If no output is specified, output to "schedule" with the same rules

		for (int i = 0; i < args.length; i++) {
			String arg = args[i];

			if (arg.equals("--noconstraints")) {
				CotiroLogic.println("Not using constraints.");
				doConstraints = false;
			} else if (arg.equals("--noconsistency")) {
				CotiroLogic.println("Not forcing arc consistency.");
				doConsistency = false;
			} else if (arg.equals("--noheuristics")) {
				CotiroLogic.println("Not using heuristics for search ordering.");
				doHeuristics = false;
			} else if (arg.equals("--nosoft")) {
				CotiroLogic.println("Not using soft constraints.");
				doSoft = false;
			} else if (arg.equals("--showstates")) {
				CotiroLogic.println("Showing all visited states.");
				showStates = true;
			} else if (arg.equals("--noconsoleoutput")) {
				CotiroLogic.println("Not printing final schedule to console.");
				consoleOutput = false;
			} else if (arg.equals("--overwrite")) {
				CotiroLogic.println("Overwriting pre-existing files.");
				overwrite = true;
			} else if (arg.equals("--output")) {
				if (i + 1 >= args.length || args[i + 1].startsWith("--")) {
					System.err
							.println("Error: Argument 'output' requires a parameter!");
					return;
				} else {
					/*outputName = "schedules"
							+ System.getProperty("file.separator")
							+ args[i + 1];*/
					outputName = args[i+1];
					i++;
					CotiroLogic.println("Outputting files to " + outputName);
				}
			} else if (arg.equals("--input")) {
				if (i + 1 >= args.length || args[i + 1].startsWith("--")) {
					System.err
							.println("Error: Argument 'input' requires a parameter!");
					return;
				} else {
					/*inputName = "data" + System.getProperty("file.separator")
							+ args[i + 1];*/
					inputName = args[i+1];
					
					// use the input file to form the output file name, in case one is not specified
					if (outputName == null) {
						outputName = args[i+1];
						
						// trim .xml
						outputName = outputName.replaceAll("\\.xml$", "");
					}
					i++;
					CotiroLogic.println("Input file: " + inputName);
				}
			} else {
				System.err.println("Unknown argument:" + arg);
				return;
			}
		}

		if (inputName == null) {
			CotiroLogic.println("ERROR! No input file specified!");
			return;
		}
		// Check for the existence of .xml, .txt, and .html files,
		// adjust the file prefix accordingly
		if (!CotiroLogic.checkFilePrefix(outputName, overwrite)) {
			return;
		}

		// Here we probably want parse and load the input XML file
		// then we can supply the XML beans to methods that would have normally
		// done parsing themselves

		try {
			cotiroInput = CotiroLogic.loadInputFile(inputName);
		} catch (XmlException e1) {
			System.err.println("Error parsing input XML");
			e1.printStackTrace();
			System.exit(-1);
		} catch (IOException e1) {
			System.err.println("Error opening input file");
			e1.printStackTrace();
			System.exit(-1);
		}

	
		
		if (cotiroInput.getEmployees() != null
				&& cotiroInput.getShifts() != null) {
			org.leedberg.cotiro.shiftbased.Scheduler.doSchedule(cotiroInput,
					outputName, inputName, doConstraints, doConsistency,
					doSoft, doHeuristics, showStates, consoleOutput);
		} else if (cotiroInput.getBuildings() != null && cotiroInput.getCourses() != null) {
			org.leedberg.cotiro.college.Scheduler.doSchedule(cotiroInput, outputName, inputName, doConstraints, doConsistency, doSoft, doHeuristics, showStates, consoleOutput);
		} else {
			CotiroLogic.println("ERROR: Input file does not seem to contain a complete scheduling problem!");
		}
	}

}
