package org.leedberg.cotiro;

import java.io.File;
import java.io.IOException;
import java.util.ArrayList;
import java.util.Collections;
import java.util.HashMap;
import java.util.Iterator;
import java.util.Stack;
import java.util.Vector;

import org.apache.xmlbeans.XmlException;
import org.leedberg.cotiro.CotiroDocument.Cotiro;

/*
Cotiro 0.6.5
College Time and Room Scheduler Using Constraint Satisfaction
Copyright (C) 2003 - 2007, Gregory G. Leedberg

This program is free software; you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation; either version 2 of the License, or
(at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program; if not, write to the Free Software
Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
*/
public class CotiroLogic {
	public enum OutputDestination {CONSOLE, STRINGARRAY }
	public static OutputDestination outputDestination = OutputDestination.CONSOLE;
	public static ArrayList stringArrayOutput;
	
	public static final String version = "Cotiro 0.7.8";
		
	public static void println(String s) {		
		if (outputDestination == OutputDestination.CONSOLE) {
			System.out.println(s);
		} else if (outputDestination == OutputDestination.STRINGARRAY) {
			if (stringArrayOutput == null) {
				stringArrayOutput = new ArrayList();
			}
			
			stringArrayOutput.add(s);
			System.err.println(s);
		}
	}
	public static void clearOutput() {
		stringArrayOutput = null;
	}
	
	public static State backTrackSearch(Stack stack, Constraints constraints, boolean showStates) {
		boolean foundAssignment = false;
		State endState = null;
		int consideredStates = 0;
		while (stack.size() > 0 && !foundAssignment) {
			State curState = (State) stack.pop();

			if (showStates) {
				CotiroLogic.println("Considering state:");
				CotiroLogic.println(curState.toString());
			}

			consideredStates++;

			if (!constraints.satisfied(curState)) {
				// Bad assignment, do nothing
				// This is not a goal state, and none of its successors
				// can be, either

				if (showStates)
					CotiroLogic.println("Rejected due to violated constraints");
			} else if (curState.allAssigned()) {
				// Satisfies constraints and is a complete assignment,
				// we are done
				foundAssignment = true;
				endState = curState;
			} else {
				// Keep looking at possible assignments
				Stack successors = generate_successors(curState);

				// Want to put successors at head of stack
				stack.addAll(successors);
			}
		}
		
		return endState;

	}
	
	public static void make1consistent(State state, Constraints constraints) {
		// We want to go through the domain of each unassigned variable,
		// and remove vlaues which violate unary constraints
		VariableList list = state.getUnassignedVars();

		// Loop through each constraint
		for (int i = 0; i < constraints.size(); i++) {
			Constraint c = constraints.getConstraint(i);
                        System.err.println("Looking at: " + c.toString());
                        
			if (c.isUnary() && !c.isSoft) {
                                // TODO: This assumes that A in the constraint is a variable!
				String id = (String) c.getVarA();
				Variable v = (Variable)list.getVariable(id);
				Vector domain = v.getDomain();

				// Try each value of the domain, and see if it violates the
				// constraint
				Iterator j = domain.iterator();
				while (j.hasNext()) {
					Value val = (Value) j.next();
					v.setValue(val);
					VariableList tempList = new VariableList();
					tempList.add(v);
					if (!c.satisfied(tempList)) {
						// This doctor violates the current unary constraint
						j.remove(); // Remove from domain
					}
				}

				v.setValue(null);
			}
		}
	}
	
	// For a given state, return the possible assignments
	public static Stack generate_successors(State state) {
		// This involves assigning every available value to all unassigned
		// variables.
		Vector unassignedVars = state.getUnassignedVars();

		Stack successors = new Stack();
		// ArrayList successors = new ArrayList();

		// for (int i=0; i < unassignedVars.size(); i++) {
		Vector availableValues = ((Variable) unassignedVars.get(0)).getDomain();

		for (int j = 0; j < availableValues.size(); j++) {
			// Create copy of the variable, assign it value, add to stack
			State newState = (State) state.clone();

			// System.err.println("Successor: " + newState.toString());
			Variable elem = (Variable) newState.getUnassignedVars()
					.remove(0);
			Value val = (Value) elem.getDomain().remove(j);
			elem.setValue(val);
			newState.getAssignedVars().add(elem);

			// successors.push(newState);
			successors.add(newState);
		}
		// }

		// We should, maybe, generate a list, then sort that list, low-to-high,
		// then
		// iterate over the list pushing each element onto the stack
		// TODO: We should sort based on the h-val of the successors --
		// the higher the h-val, the higher priority it should get (most
		// constrained)

		return successors;
	}
	
	// Calculates heuristic values for variable ordering, based on
	// degree of constraint for each variable
	public static void calculateHvals(State state, Constraints constraints) {
		HashMap hvals = new HashMap();

		for (int i = 0; i < constraints.size(); i++) {
			Constraint c = constraints.getConstraint(i);

			Object c1 = c.getVarA();
			Object c2 = null;
			if (!c.isUnary()) {
				c2 = (Object) c.getVarB();
			}

			// Has to at least have c1
			Integer curCount = (Integer) hvals.get(c1);
			if (curCount == null) {
				// First occurrence
				curCount = new Integer(1);
				hvals.put(c1, curCount);
			} else {
				// already exists
				int cc = curCount.intValue();
				cc++;
				curCount = new Integer(cc);
				hvals.put(c1, curCount);

			}

			// If constraint is binary, then we need to work on c2 as well
			if (!c.isUnary()) {
				curCount = (Integer) hvals.get(c2);
				if (curCount == null) {
					// First occurrence
					curCount = new Integer(1);
					hvals.put(c2, curCount);
				} else {
					// already exists
					int cc = curCount.intValue();
					cc++;
					curCount = new Integer(cc);
					hvals.put(c2, curCount);

				}

			}
		}

		// Now, we have a hashmap of all h-vals. now we just need to
		// register them with the variables themselves
		Vector bnl = state.getUnassignedVars();

		Iterator i = bnl.iterator();
		while (i.hasNext()) {
			Variable bn = (Variable) i.next();
			Integer h_val = (Integer) hvals.get(bn.getUniqueID());

			if (h_val == null) {
				// course did not appear in ANY constraints
				bn.setHval(0);
			} else {
				bn.setHval(h_val.intValue());
			}
		}
	}
	



	// Order the values for in the domain of each variable in this state
	// according to the number of soft-constraints the values satisfy.
	public static void orderValues(State state, Constraints constraints) {
		// We assume that this is just an inital state, so all variables
		// appear in unassignedVars
		Iterator i = state.unassignedVars.iterator();

		while (i.hasNext()) {
			Variable bn = (Variable) i.next();

			// Sort each value in the domain according to how many
			// soft constraints it satisfies
			CotiroLogic.sortDomain(bn, constraints);
		}
	}
	
	// Sorts c's domain of values according to how many
	// soft (unary) constraints are satisfied
	public static void sortDomain(Variable bn, Constraints constraints) {
		// First find constraints which are both unary, soft,
		// and only involve this variable
		Constraints applicable = new Constraints();

		for (int i = 0; i < constraints.size(); i++) {
			Constraint t = constraints.getConstraint(i);

			// We use t.A.equals because t.A is a STRING, not a pointer
			// to the coursesection object!
			if (t.isSoft && t.isUnary() && t.A.equals(bn.getUniqueID())) {
				applicable.addConstraint(t);
			}
		}

		// If there are no applicable constraints, then there's no need
		// to continue
		if (applicable.size() == 0)
			return;

		// Now, iterate over every value in the domain for this
		// course and see how many soft unary constraints are
		// satisfied
		Iterator i = bn.getDomain().iterator();

		while (i.hasNext()) {
			// can use Collections.sort to do the work.

			// First, let's create a state in which this value
			// is the assigned value.
			Value d = (Value) i.next();

			// Assign this value
			bn.setValue(d);

			VariableList assignedVars = new VariableList();
			VariableList unassignedVars = new VariableList();

			// Make this coursesection the only assigned variable in
			// the state
			assignedVars.add(bn);
			State state = new State(assignedVars, unassignedVars);

			// Determine how many applicable constraints are
			// satisifed with this assignment

			// NOTE: I believe that there is only one instance of each
			// TimeRoom object, so this h_val field will be overwritten
			// for each CourseSection object. That won't break this
			// algorithm, however, since we only need this h_val
			// for sorting at the end of this function. Once sorted
			// within the domain Vector for this CourseSection, we don't
			// need to remember the calculated h-val again, so it's okay
			// to keep overwriting it. Worth noting though in case
			// needs change in the future!!
			d.setHval(applicable.numSatisfied(state));

			// Reset assignment
			bn.setValue(null);
		}

		// Now we have calculated h-vals for each value in the domain.
		// All we need to do is sort.
		// We've defined TimeRoom objects to be sorted according to h-val
		Collections.sort(bn.getDomain());

		return;
	}
	

	static public  Cotiro loadInputFile(String inputFile) throws XmlException,
			IOException {
		CotiroDocument doc = CotiroDocument.Factory.parse(new File(inputFile));
		return doc.getCotiro();
	}
	
	// return TRUE if prefix is okay, false if otherwise
	public static boolean checkFilePrefix(String output, boolean overwrite) {
		boolean okay = true;

		File f1 = new File(output + ".xml");
		File f2 = new File(output + ".txt");
		File f3 = new File(output + ".html");

		if (f1.exists() && !overwrite) {
			CotiroLogic.println("ERROR: The specified output file already exists: "
							+ f1.getAbsolutePath());
			okay = false;
		} else if (f1.exists()) {
			CotiroLogic.println("WARNING: The specified output file already exists: "
							+ f1.getAbsolutePath());
		}

		if (f2.exists() && !overwrite) {
			CotiroLogic.println("ERROR: The specified output file already exists: "
							+ f2.getAbsolutePath());
			okay = false;
		} else if (f2.exists()) {
			CotiroLogic.println("WARNING: The specified output file already exists: "
							+ f2.getAbsolutePath());

		}

		if (f3.exists() && !overwrite) {
			CotiroLogic.println("ERROR: The specified output file already exists: "
							+ f3.getAbsolutePath());
			okay = false;
		} else if (f3.exists()) {
			CotiroLogic.println("WARNING: The specified output file already exists: "
							+ f3.getAbsolutePath());
		}

		if (!okay) {
			CotiroLogic.println("");
			CotiroLogic.println("The currently specified output file prefix is: "
							+ output);
			CotiroLogic.println("");
			CotiroLogic.println("The files that would be produced using this prefix are: ");
			CotiroLogic.println(f1.getAbsolutePath());
			CotiroLogic.println(f2.getAbsolutePath());
			CotiroLogic.println(f3.getAbsolutePath());
			CotiroLogic.println("");
			CotiroLogic.println("One or more of these files already exist.");
			CotiroLogic.println("");
			CotiroLogic.println("Please specify a unique output file prefix using --output,");
			CotiroLogic.println("or use the --overwrite argument to overwrite existing files.");

			return false;

		}
		else {
			return true;
		}

	}

}
