package org.leedberg.cotiro.college;

import java.io.File;
import java.util.Collections;
import java.util.HashMap;
import java.util.Iterator;
import java.util.Properties;
import java.util.Set;
import java.util.Stack;
import java.util.Vector;

import org.apache.xmlbeans.XmlOptions;
import org.leedberg.cotiro.Constraint;
import org.leedberg.cotiro.Constraints;
import org.leedberg.cotiro.CotiroLogic;
import org.leedberg.cotiro.Days;
import org.leedberg.cotiro.SoftSetting;
import org.leedberg.cotiro.State;
import org.leedberg.cotiro.Time;
import org.leedberg.cotiro.TimeSpan;
import org.leedberg.cotiro.CotiroDocument.Cotiro;
import org.leedberg.cotiro.CotiroDocument.Cotiro.Buildings;
import org.leedberg.cotiro.CotiroDocument.Cotiro.Courses;
import org.leedberg.cotiro.CotiroDocument.Cotiro.Schedules;
import org.leedberg.cotiro.CotiroDocument.Cotiro.Teachers;
import org.leedberg.cotiro.CotiroDocument.Cotiro.Schedules.Schedule;
import org.leedberg.cotiro.CotiroDocument.Cotiro.Teachers.Teacher;
import org.leedberg.cotiro.CotiroDocument.Cotiro.Teachers.Teacher.Properties.Property;
import org.leedberg.cotiro.college.constraints.BuildingConstraint;
import org.leedberg.cotiro.college.constraints.InSpanConstraint;
import org.leedberg.cotiro.college.constraints.NoOverlapConstraint;
import org.leedberg.cotiro.college.constraints.NotSameRoomConstraint;
import org.leedberg.cotiro.college.constraints.SizeConstraint;

/*
Cotiro 0.6.5
College Time and Room Scheduler Using Constraint Satisfaction
Copyright (C) 2003 - 2007, Gregory G. Leedberg

This program is free software; you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation; either version 2 of the License, or
(at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program; if not, write to the Free Software
Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
*/

/*
 * Created on Nov 3, 2003
 *
 */

public class Scheduler {

	// Default class length 50 minutes
	protected static int classDuration = 50;

	// Default break between classes 10 minutes
	protected static int classBreak = 10;

	protected static Vector allPossibleTimeSpans = new Vector();
        
        static HashMap<String, java.util.Properties> teacherProperties = new HashMap<String, java.util.Properties>();

	/**
	 * 
	 */
	public Scheduler() {
		super();
	}

	public static void doSchedule(Cotiro cotiroInput, String outputName, String inputName, boolean doConstraints, boolean doConsistency, boolean doSoft, boolean doHeuristics, boolean showStates, boolean consoleOutput) {
		
		loadScheduleSettings(cotiroInput.getSchedules());

		State initial_state = loadInitialState(cotiroInput.getCourses(), cotiroInput.getBuildings(), cotiroInput.getTeachers());
		Constraints constraints = new Constraints();
		Stack stack = new Stack();
		boolean foundAssignment = false;
		State endState = new State();

		if (doConstraints) {
			constraints = buildConstraints(initial_state, doSoft, cotiroInput.getTeachers());
			CotiroLogic.println("Constraints: ");
			CotiroLogic.println(constraints.toString());
		}

		CotiroLogic.println("-----------------------------------------");
		CotiroLogic.println("**********PRE-PROCESSING STEPS:**********");

		if (doConsistency && doConstraints) {
			CotiroLogic.println("Forcing 1 consistency...");
			CotiroLogic.make1consistent(initial_state, constraints);
			CotiroLogic.println("done");
		}

		if (doHeuristics && doConstraints) {
			CotiroLogic.println("Calculating h-vals...");
			CotiroLogic.calculateHvals(initial_state, constraints);
			CotiroLogic.println("done");
		}

		if (doConstraints && doSoft) {
			CotiroLogic.println("Ordering values for soft constraints...");
			CotiroLogic.orderValues(initial_state, constraints);
			CotiroLogic.println("done");
		}

		// At this point, we actually want to REMOVE soft constraints
		// from the list of constraints, so that states that violate them
		// are allowed. We have already done everything we need with those
		// soft states (value ordering).
		if (doConstraints && doSoft) {
			CotiroLogic.println("Removing soft constraints from system...");
			for (int i = 0; i < constraints.size(); i++) {
				if (((Constraint) constraints.getConstraint(i)).isSoft) {
					constraints.removeConstraint(i);
					i--;
				}
			}
		}

		CotiroLogic.println("-----------------------------------------");

		stack.push(initial_state);

		// Back-tracking search (depth-first search)
		int consideredStates = 0;
		while (stack.size() > 0 && !foundAssignment) {
			State curState = (State) stack.pop();

			if (showStates) {
				CotiroLogic.println("Considering state:");
				CotiroLogic.println(curState.toString());
			}

			consideredStates++;

			if (!constraints.satisfied(curState)) {
				// Bad assignment, do nothing
				// This is not a goal state, and none of its successors
				// can be, either

				if (showStates)
					CotiroLogic.println("Rejected due to violated constraints");
			} else if (curState.allAssigned()) {
				// Satisfies constraints and is a complete assignment,
				// we are done
				foundAssignment = true;
				endState = curState;
			} else {
				// Keep looking at possible assignments
				Stack successors = CotiroLogic.generate_successors(curState);

				// Want to put successors at head of stack
				stack.addAll(successors);
			}
		}
		if (foundAssignment) {
			// CotiroLogic.println(endState.getAssignedVars().toString());
			// CotiroLogic.println(endState.getAssignedVars().getXmlBean()
			// .toString());

			try {
				// Sort the schedule based on course name, this should
				// help with the readability of the output file.
				
				Collections.sort(endState.getAssignedVars());

				((CourseSectionList)endState.getAssignedVars()).getXmlBean().save(
						new File(outputName + ".xml"),
						new XmlOptions().setUseDefaultNamespace());

				// Output text and HTML
				XmlConverter.main(new String[] { "-text", outputName + ".xml",
						outputName + ".txt" });
				XmlConverter.main(new String[] { "-html", outputName + ".xml",
						outputName + ".html" });

				if (consoleOutput) {
					CotiroLogic.println(XmlConverter.convertToText(new File(
							outputName + ".xml")));
				}
			} catch (Exception e) {
				System.err.println(e.toString());
				e.printStackTrace();
			}

		} else {
			CotiroLogic.println("Impossible to make legal assignments!");
		}

		CotiroLogic.println("\n**Traversed " + consideredStates + " states");
	}

	static private void loadScheduleSettings(Schedules schedules) {
		// ScheduleDocument scheduleDoc = null;
		// SchedulesDocument schedulesDoc = null;

		// Bind the incoming XML to an XMLBeans type.
		// scheduleDoc = ScheduleDocument.Factory.parse(new File(
		// "data/schedule.xml"));
		// schedulesDoc = SchedulesDocument.Factory.parse(new File(
		// "data/schedules.xml"));

		Schedule[] arr = schedules.getScheduleArray();

		for (int i = 0; i < arr.length; i++) {
			Schedule schedule = arr[i];

			classDuration = schedule.getCourseDuration().intValue();

			classBreak = schedule.getCourseBreak().intValue();

			TimeSpan span;

			// Default day span 8:10 - 20:00pm
			TimeSpan day = new TimeSpan(new Time(8, 10), new Time(20, 0),
					new Days());

			day = new TimeSpan(new Time(schedule.getDayStart()), new Time(
					schedule.getDayEnd()), new Days());

			Time startTime = (Time) day.getStart().clone();
			Time endTime = (Time) day.getStart().clone();

			endTime.addMinutes(classDuration);
			span = new TimeSpan(startTime, endTime, new Days());

			if (day.encompasses(span)) {
				Scheduler.allPossibleTimeSpans.add(span);
			}

			while (day.encompasses(span)) {
				startTime = (Time) endTime.clone();
				endTime = (Time) startTime.clone();

				startTime.addMinutes(classBreak);
				endTime.addMinutes(classDuration + classBreak);

				span = new TimeSpan(startTime, endTime, new Days(schedule
						.getDays()));

				if (day.encompasses(span)) {
					Scheduler.allPossibleTimeSpans.add(span);
				}
			}

		}

	}

	// loadInitial state loads in the state info from the relevant files and
	// returns the
	// resulting initial state
	static private State loadInitialState(Courses courses, Buildings buildings, Teachers teachers) {
		CourseSectionList unassignedVars = new CourseSectionList();
		Vector availableValues = new Vector();
		CourseSectionList assignedVars = new CourseSectionList(); // Will be
		// left
		// empty

		// Read in courses and build up the unassignedVariables vector
		
		// Iterate over each course
		for (int i = 0; i < courses.getCourseArray().length; i++) {
			CourseSection course;
			String courseName;
			String instructorName;
                        
			int size;

			courseName = courses.getCourseArray()[i].getTitle();
			instructorName = courses.getCourseArray()[i].getInstructor();
                    
                        
                        
			size = courses.getCourseArray()[i].getSize().intValue();

			course = new CourseSection(courseName, instructorName, size);
			unassignedVars.add(course);

		}

		// Read in the rooms and generate the availableValues vector
		// Rooms file

		// Iterate over each building
		for (int i = 0; i < buildings.getBuildingArray().length; i++) {
			String building;
			building = buildings.getBuildingArray()[i].getName();

			// Iterate over each room
			for (int j = 0; j < buildings.getBuildingArray()[i].getRoomArray().length; j++) {
				String room;
				int size;

				room = buildings.getBuildingArray()[i].getRoomArray()[j]
						.getNum();
				size = buildings.getBuildingArray()[i].getRoomArray()[j]
						.getSize().intValue();

				// Need to create a block of time for this room for each class
				// period
				// during the day

				TimeRoom timeRoom;

				Iterator iter = Scheduler.allPossibleTimeSpans.iterator();

				while (iter.hasNext()) {
					TimeSpan ts = (TimeSpan) iter.next();

					timeRoom = new TimeRoom(building, room, size, ts);
					availableValues.add(timeRoom);

				}
			}
		}
                
                // Need to set up teacher properties
                for (Teacher t : teachers.getTeacherArray()) {
                    Properties inmemProps = Scheduler.teacherProperties.get(t.getName());
                    
                    if (inmemProps == null) {
                        inmemProps = new Properties();
                        Scheduler.teacherProperties.put(t.getName(), inmemProps);
                    }
                    
                    org.leedberg.cotiro.CotiroDocument.Cotiro.Teachers.Teacher.Properties xmlprops = t.getProperties();
                    // not sure if it could ever be null, but just to be safe...
                    if (xmlprops != null) {
                        for (Property p : xmlprops.getPropertyArray()) {
                            inmemProps.setProperty(p.getName(), p.getValue());
                        }
                        
                    }
                }

		CotiroLogic.println(Integer.toString(availableValues.size()));
		// Start each variable with a domain of all possible values
		for (int i = 0; i < unassignedVars.size(); i++) {
			CourseSection course = (CourseSection) unassignedVars.get(i);
			course.domain = (Vector) availableValues.clone();
		}
		return new State(assignedVars, unassignedVars);
	}

	// Build constraints based on implicit restraints from class times and
	// teachers,
	// and load explicit constraints from teachers file, such as teacher
	// buildings.
	private static Constraints buildConstraints(State initial_state,
			boolean doSoft, Teachers teachers_in) {
		Constraints constraints = new Constraints();
		CourseSectionList list = (CourseSectionList)initial_state.getUnassignedVars();
		HashMap teacherBuildings = new HashMap();
		HashMap teacherSpans = new HashMap();

		// Read in teacher building preferences


		for (int i = 0; i < teachers_in.getTeacherArray().length; i++) {
			String teacherName = teachers_in.getTeacherArray()[i].getName();
			String building = teachers_in.getTeacherArray()[i].getBuilding();

			String startString = teachers_in.getTeacherArray()[i]
					.getStartTime();
			String endString = teachers_in.getTeacherArray()[i].getEndTime();

			if (building != null) {
				teacherBuildings.put(teacherName, building);
			}

			if (startString != null && endString != null) {
				Time start = new Time(startString);
				Time end = new Time(endString);

				TimeSpan available = new TimeSpan(start, end, new Days());
				teacherSpans.put(teacherName, available);
			}

		}
		// Generate constraints for class / room size
		for (int i = 0; i < list.size(); i++) {
			CourseSection course = (CourseSection) list.get(i);
			String courseName = course.getCourse();

			Constraint constraint = new SizeConstraint(courseName,
					new SoftSetting(doSoft, true, false));
			constraints.addConstraint(constraint);
		}

		// Generate constraints for teacher time conflicts
		HashMap teachers = new HashMap();
		for (int i = 0; i < list.size(); i++) {
			CourseSection course = (CourseSection) list.get(i);
			String teacherName = course.getInstructor();
			CourseSectionList teacherCourses = (CourseSectionList) teachers
					.get(teacherName);

			if (teacherCourses == null) {
				CourseSectionList l = new CourseSectionList();
				l.add(course);
				teachers.put(teacherName, l);
			} else {
				teacherCourses.add(course);
			}
		}

		// Now we have lists of courses taught by each teacher
		Set keys = teachers.keySet();

		// Establish time conflict constraints
		Iterator i = keys.iterator();
		while (i.hasNext()) {
			String teacherName = (String) i.next();
			CourseSectionList l = (CourseSectionList) teachers.get(teacherName);
			String building = (String) teacherBuildings.get(teacherName);
			TimeSpan span = (TimeSpan) teacherSpans.get(teacherName);

			for (int j = 0; j < l.size(); j++) {
				CourseSection course = (CourseSection) l.get(j);
				String courseName = course.getCourse();

				// Want to create a building constraint as well
				if (building != null) {
					Constraint c = new BuildingConstraint(courseName, building,
							new SoftSetting(doSoft, true, false));
					constraints.addConstraint(c);
				}

				// And time span constraint
				if (span != null) {
					Constraint c = new InSpanConstraint(courseName, span,
							new SoftSetting(doSoft, true, false));
					constraints.addConstraint(c);
				}

				// Now, create all time conflict constraints
				for (int k = j + 1; k < l.size(); k++) {
					CourseSection course2 = (CourseSection) l.get(k);
					String courseName2 = course2.getCourse();

					Constraint c = new NoOverlapConstraint(courseName,
							courseName2, new SoftSetting(doSoft, true, false));
					constraints.addConstraint(c);
				}
			}
		}

		// Now, generate the constraints that say no two classes can be in the
		// same
		// room at the same time
		for (int k = 0; k < list.size(); k++) {
			String course1 = ((CourseSection) list.get(k)).getCourse();
			for (int j = k + 1; j < list.size(); j++) {
				String course2 = ((CourseSection) list.get(j)).getCourse();
				Constraint c = new NotSameRoomConstraint(course1, course2,
						new SoftSetting(doSoft, true, false));
				constraints.addConstraint(c);
			}
		}
		return constraints;
	}
}
